import numpy as np
import pandas as pd
import pytest
import xarray as xr

from mlair.helpers.statistics import standardise, standardise_inverse, standardise_apply, centre, centre_inverse, \
    centre_apply, apply_inverse_transformation, min_max, min_max_inverse, min_max_apply, log, log_inverse, log_apply

lazy = pytest.lazy_fixture


@pytest.fixture(scope='module')
def input_data():
    return np.array([np.random.normal(2, 2, 3000),
                     np.random.normal(-5, 3, 3000),
                     np.random.normal(10, 1, 3000)]).T


@pytest.fixture(scope='module')
def input_data_gamma():
    return np.array([np.random.gamma(2, 2, 3000),
                     np.random.gamma(3, 3, 3000),
                     np.random.gamma(1, 1, 3000)]).T


@pytest.fixture(scope='module')
def pandas(input_data):
    return pd.DataFrame(input_data)


@pytest.fixture(scope='module')
def pandas_gamma(input_data_gamma):
    return pd.DataFrame(input_data_gamma)


@pytest.fixture(scope='module')
def pd_mean():
    return [2, 10, 3]


@pytest.fixture(scope='module')
def pd_std():
    return [3, 2, 3]


@pytest.fixture(scope='module')
def pd_min():
    return pd.Series([2, -10, -3])


@pytest.fixture(scope='module')
def pd_max():
    return pd.Series([3, 2, 3])


@pytest.fixture(scope='module')
def xarray(input_data):
    shape = input_data.shape
    coords = {'index': range(shape[0]), 'value': range(shape[1])}
    return xr.DataArray(input_data, coords=coords, dims=coords.keys())


@pytest.fixture(scope='module')
def xarray_gamma(input_data_gamma):
    shape = input_data_gamma.shape
    coords = {'index': range(shape[0]), 'value': range(shape[1])}
    return xr.DataArray(input_data_gamma, coords=coords, dims=coords.keys())


@pytest.fixture(scope='module')
def xr_mean(input_data):
    return xr.DataArray([2, 10, 3], coords={'value': range(3)}, dims=['value'])


@pytest.fixture(scope='module')
def xr_std(input_data):
    return xr.DataArray([3, 2, 3], coords={'value': range(3)}, dims=['value'])


@pytest.fixture(scope='module')
def xr_min(input_data):
    return xr.DataArray([2, -10, -3], coords={'value': range(3)}, dims=['value'])


@pytest.fixture(scope='module')
def xr_max(input_data):
    return xr.DataArray([3, 2, 3], coords={'value': range(3)}, dims=['value'])


class TestStandardise:

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_standardise(self, data_orig, dim):
        data, opts = standardise(data_orig, dim)
        assert np.testing.assert_almost_equal(opts["mean"], [2, -5, 10], decimal=1) is None
        assert np.testing.assert_almost_equal(opts["std"], [2, 3, 1], decimal=1) is None
        assert np.testing.assert_almost_equal(data.mean(dim), [0, 0, 0]) is None
        assert np.testing.assert_almost_equal(data.std(dim), [1, 1, 1]) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_standardise_inverse(self, data_orig, dim):
        data, opts = standardise(data_orig, dim)
        data_recovered = standardise_inverse(data, opts["mean"], opts["std"])
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_apply_standardise_inverse(self, data_orig, dim):
        data, opts = standardise(data_orig, dim)
        data_recovered = apply_inverse_transformation(data, **opts)
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, mean, std, dim', [(lazy('pandas'), lazy('pd_mean'), lazy('pd_std'), 0),
                                                           (lazy('xarray'), lazy('xr_mean'), lazy('xr_std'), 'index')])
    def test_standardise_apply(self, data_orig, mean, std, dim):
        data = standardise_apply(data_orig, mean, std)
        mean_expected = (np.array([2, -5, 10]) - np.array([2, 10, 3])) / np.array([3, 2, 3])
        std_expected = np.array([2, 3, 1]) / np.array([3, 2, 3])
        assert np.testing.assert_almost_equal(data.mean(dim), mean_expected, decimal=1) is None
        assert np.testing.assert_almost_equal(data.std(dim), std_expected, decimal=1) is None


class TestCentre:

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_centre(self, data_orig, dim):
        data, opts = centre(data_orig, dim)
        assert np.testing.assert_almost_equal(opts["mean"], [2, -5, 10], decimal=1) is None
        assert np.testing.assert_almost_equal(data.mean(dim), [0, 0, 0]) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_centre_inverse(self, data_orig, dim):
        data, opts = centre(data_orig, dim)
        data_recovered = centre_inverse(data, opts["mean"])
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_apply_centre_inverse(self, data_orig, dim):
        data, opts = centre(data_orig, dim)
        data_recovered = apply_inverse_transformation(data, **opts)
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, mean, dim', [(lazy('pandas'), lazy('pd_mean'), 0),
                                                      (lazy('xarray'), lazy('xr_mean'), 'index')])
    def test_centre_apply(self, data_orig, mean, dim):
        data = centre_apply(data_orig, mean)
        mean_expected = np.array([2, -5, 10]) - np.array([2, 10, 3])
        assert np.testing.assert_almost_equal(data.mean(dim), mean_expected, decimal=1) is None


class TestMinMax:

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_min_max(self, data_orig, dim):
        data, opts = min_max(data_orig, dim)
        max_expected = data_orig.max(dim)
        min_expected = data_orig.min(dim)
        assert np.testing.assert_array_almost_equal(opts["max"], max_expected, decimal=1) is None
        assert np.testing.assert_array_almost_equal(opts["min"], min_expected, decimal=1) is None
        assert np.testing.assert_almost_equal(data.max(dim), [1, 1, 1]) is None
        assert np.testing.assert_almost_equal(data.min(dim), [0, 0, 0]) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_min_max_inverse(self, data_orig, dim):
        data, opts = min_max(data_orig, dim)
        data_recovered = min_max_inverse(data, opts["min"], opts["max"])
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_apply_min_max_inverse(self, data_orig, dim):
        data, opts = min_max(data_orig, dim)
        data_recovered = apply_inverse_transformation(data, **opts)
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, dmin, dmax, dim', [(lazy('pandas'), lazy('pd_min'), lazy('pd_max'), 0),
                                                            (lazy('xarray'), lazy('xr_min'), lazy('xr_max'), 'index')])
    def test_min_max_apply(self, data_orig, dmin, dmax, dim):
        data = min_max_apply(data_orig, dmin, dmax)
        min_expected = (data_orig.min(dim) - dmin) / (dmax - dmin)
        max_expected = (data_orig.max(dim) - dmin) / (dmax - dmin)
        assert np.testing.assert_array_almost_equal(data.min(dim), min_expected, decimal=1) is None
        assert np.testing.assert_array_almost_equal(data.max(dim), max_expected, decimal=1) is None


class TestLog:

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas_gamma'), 0),
                                                (lazy('xarray_gamma'), 'index')])
    def test_standardise(self, data_orig, dim):
        data, opts = log(data_orig, dim)
        assert {"method", "mean", "std"} == opts.keys()
        assert opts["method"] == "log"
        assert np.testing.assert_almost_equal(data.mean(dim), [0, 0, 0]) is None
        assert np.testing.assert_almost_equal(data.std(dim), [1, 1, 1]) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas_gamma'), 0),
                                                (lazy('xarray_gamma'), 'index')])
    def test_standardise_inverse(self, data_orig, dim):
        data, opts = log(data_orig, dim)
        data_recovered = log_inverse(data, opts["mean"], opts["std"])
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas_gamma'), 0),
                                                (lazy('xarray_gamma'), 'index')])
    def test_apply_standardise_inverse(self, data_orig, dim):
        data, opts = log(data_orig, dim)
        data_recovered = apply_inverse_transformation(data, **opts)
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_standardise_apply(self, data_orig, dim):
        data_ref, opts = log(data_orig, dim)
        data_test = log_apply(data_orig, opts["mean"], opts["std"])
        assert np.testing.assert_array_almost_equal(data_ref, data_test) is None
